/* Copyright (c) 2007 Michele Bini
 * 
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA */


#include <gmp.h>
#include "curve25519.h"

static curve25519key_t p25519 = { 0xffffffed, 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff, 0x7fffffff };
static curve25519key_t zerocmp = { 0, 0, 0, 0, 0, 0, 0, 0 };
static curve25519key_t onecmp = { 1, 0, 0, 0, 0, 0, 0, 0 };

extern int
getbit(curve25519key_t *x, unsigned int n) {
  unsigned int d = (sizeof(zerocmp[0])*8);
  return (x[0][n / d] >> (n % d)) & 1;
}

extern void
setbit(curve25519key_t *x, unsigned int n, int v) {
  unsigned int d = (sizeof(zerocmp[0])*8);
  unsigned int i = n / d;
  mp_limb_t l = x[0][i];
  if (v) {
    l |= 1<<(n % d);
  } else {
    l &= ~((mp_limb_t)(1<<(n % d)));
  }
  x[0][i] = l;
}

extern unsigned int
getbyte(curve25519key_t *x, unsigned int n) {
  unsigned int d = (sizeof(zerocmp[0])*8);
  n *= 8;
  return (x[0][n / d] >> (n % d)) & 0xff;
}

extern void
setbyte(curve25519key_t *x, unsigned int n, unsigned int v) {
  unsigned int d = (sizeof(zerocmp[0])*8);
  unsigned int i = n / d;
  mp_limb_t l = x[0][i];
  n *= 8;
  l &= ~((mp_limb_t)(0xff<<(n % d)));
  l |= v<<(n % d);
  x[0][i] = l;
}

#if 0
static void
tracev(char*m, curve25519key_t *x) {
  char s[(C25519BITS/4)+2];
  base32_encode(s, x);
  fprintf(stderr, "%s: %s\n", m, s);
}
#endif

static void
copycmp(curve25519key_t *n, curve25519key_t *x) {
  int c;
  for (c = 0; c < C25519N; c++) {
    n[0][c] = x[0][c];
  }
}

static
int zeromodp(curve25519key_t *x) {
  return (mpn_cmp((mp_limb_t*)x, (mp_limb_t*)&zerocmp, C25519N) == 0);
}

static
void addmodp(curve25519key_t *a, curve25519key_t *b) {
  mpn_add_n((mp_limb_t*)a, (mp_limb_t*)a, (mp_limb_t*)b, C25519N);
  if (mpn_cmp((mp_limb_t*)a, (mp_limb_t*)&p25519, C25519N) >= 0) {
    mpn_sub_n((mp_limb_t*)a, (mp_limb_t*)a, (mp_limb_t*)&p25519, C25519N);
  }
}

static
void submodp(curve25519key_t *a, curve25519key_t *b) {
  if (mpn_cmp((mp_limb_t*)b, (mp_limb_t*)a, C25519N) > 0) {
    mpn_add_n((mp_limb_t*)a, (mp_limb_t*)&p25519, (mp_limb_t*)a, C25519N);
  }
  mpn_sub_n((mp_limb_t*)a, (mp_limb_t*)a, (mp_limb_t*)b, C25519N);
}

static void
mulmodp(curve25519key_t *a, curve25519key_t *b) {
  mp_limb_t d[C25519N*2]; mp_limb_t r[C25519N+1];
  mpn_mul(d, (mp_limb_t*)a, C25519N, (mp_limb_t*)b, C25519N);
  mpn_tdiv_qr(r, (mp_limb_t*)a, 0, d, C25519N*2, (mp_limb_t*)&p25519, C25519N);
}

static void
sqrmodp(curve25519key_t *a) {
  mulmodp(a, a);
}

static void
invmodp(curve25519key_t *a) {
  /* a = a ** (p-2)
     0111 + (1111) x 7 + (1111) x (8*6) + (1111) x 6 + 1110 + 1011
     0 . 1 x (3 + 4*7 + 4 * 8*6 + 4*6 + 3) . 0 . 1011
     0 . 1 x (250) . 0 . 1011
  */
  curve25519key_t c; copycmp(&c, a);
  int i = 250;
  while (--i) {
    sqrmodp(a); mulmodp(a, &c);
  }
  sqrmodp(a);
  sqrmodp(a); mulmodp(a, &c);
  sqrmodp(a);
  sqrmodp(a); mulmodp(a, &c);
  sqrmodp(a); mulmodp(a, &c);
}

static mp_limb_t asmall = 121665; /* (486662 - 2) / 4; */

static
void mulasmall(curve25519key_t *a) {
  mp_limb_t d[C25519N+1]; mp_limb_t r[2];
  //tracev("mulasmall a", a);
  d[C25519N] = mpn_mul_1(d, (mp_limb_t*)a, C25519N, asmall);
  //tracev("mulasmall d", d);
  //tracev("mulasmall a", a);
  mpn_tdiv_qr(r, (mp_limb_t*)a, 0, d, C25519N+1, (mp_limb_t*)&p25519, C25519N);
  //tracev("mulasmall a", a);
}

static
void dbl(curve25519key_t *x2, curve25519key_t *z2, curve25519key_t *x, curve25519key_t *z) {
  curve25519key_t m, n, o;
  //tracev("dbl x", x);
  //tracev("dbl z", z);
  copycmp(&m, x); addmodp(&m, z); sqrmodp(&m);
  //tracev("dbl m", &m);
  copycmp(&n, x); submodp(&n, z); sqrmodp(&n);
  //tracev("dbl n", &n);
  copycmp(&o, &m); submodp(&o, &n);
  //tracev("dbl o", &o);
  copycmp(x2, &n); mulmodp(x2, &m);
  copycmp(z2, &o); mulasmall(z2); addmodp(z2, &m); mulmodp(z2, &o);
  //tracev("dbl x2", x2);
  //tracev("dbl z2", z2);
}

static
void sum(curve25519key_t *x3, curve25519key_t *z3, curve25519key_t *x, curve25519key_t *z, curve25519key_t *xp, curve25519key_t *zp, curve25519key_t *x1) {
  curve25519key_t k, l, p, q;
  //tracev("sum x", x);
  //tracev("sum z", z);
  copycmp(&p, x); submodp(&p, z); copycmp(&k, xp); addmodp(&k, zp); mulmodp(&p, &k);
  copycmp(&q, x); addmodp(&q, z); copycmp(&l, xp); submodp(&l, zp); mulmodp(&q, &l);
  //tracev("sum p", &p);
  //tracev("sum q", &q);
  copycmp(x3, &p); addmodp(x3, &q); sqrmodp(x3);
  copycmp(z3, &p); submodp(z3, &q); sqrmodp(z3); mulmodp(z3, x1);
}

extern void
curve25519(curve25519key_t *r, curve25519key_t *f, curve25519key_t *c) {
  curve25519key_t x1, xa, za, x, z;
  //tracev("f", f);
  if (zeromodp(f)) {
    copycmp(r, &zerocmp);
    return;
  }
  copycmp(&x1, c);
  //tracev("c", c);
  //tracev("x1", x1);
  dbl(&xa, &za, &x1, &onecmp);
  //tracev("xa", &xa);
  //tracev("za", &za);
  copycmp(&x, &x1);
  copycmp(&z, &onecmp);

  int n = C25519BITS-1;

  while (getbit(f, n) == 0) {
    n--;
  }
  n--;

  while (n >= 0) {
    curve25519key_t nx, nz, nxa, nza;
    int b = getbit(f, n);
    //fprintf(stderr, "b: %d\n", b);
    if (b == 0) {
      dbl(&nx, &nz, &x, &z);
      sum(&nxa, &nza, &xa, &za, &x, &z, &x1);
    } else {
      sum(&nx, &nz, &xa, &za, &x, &z, &x1);
      dbl(&nxa, &nza, &xa, &za);
    }
    copycmp(&x, &nx); copycmp(&z, &nz); copycmp(&xa, &nxa); copycmp(&za, &nza);
    //tracev("xn", &x);
    //tracev("zn", &z);
    //tracev("xa", &xa);
    //tracev("za", &za);
    n--;
  }

  //tracev("x", &x);
  //tracev("z", &z);
  invmodp(&z);
  //tracev("1/z", &z);
  mulmodp(&x, &z);
  copycmp(r, &x);
}
